//###############################################################################
//#                                                                             #
//# AX81 computer with ATmega1284P                                              #
//# copyright (c) 2010-2012 Joerg Wolfram (joerg@jcwolfram.de)                  #
//#                                                                             #
//#                                                                             #
//# This program is free software; you can redistribute it and/or               #
//# modify it under the terms of the GNU General Public License as              #
//# published by the Free Software Foundation; either version 3 of              #
//# the License, or (at your option) any later version.                         #
//#                                                                             #
//# This library is distributed in the hope that it will be useful,             #
//# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
//# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU            #
//# General Public License for more details.                                    #
//#                                                                             #
//# You should have received a copy of the GNU General Public                   #
//# License along with this library; if not, write to the                       #
//# Free Software Foundation, Inc., 59 Temple Place - Suite 330,                #
//# Boston, MA 02111-1307, USA.                                                 #
//#                                                                             #
//###############################################################################

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>

#include "ax81-tools.h"


int main(int argc, char **argv)
{
    unsigned char *thead;
    unsigned char *tprog;
    FILE * imagefile;
    char filename[1000];
    char filetype[10];
    int tape, i, rstat;
    int maxtapes[4];
    int this_image;
    const char *imagefile_name = NULL;

    thead = malloc(TAPE_HEADER_LENGTH);
    tprog = malloc(TAPE_FILE_LENGTH);

    memset(tprog, 255, TAPE_FILE_LENGTH);
    memset(thead, 255, TAPE_HEADER_LENGTH);

    thead[0] = 0xa6;
    thead[1] = 0xbd;
    thead[2] = 0x26;
    thead[3] = 0x3d;
    thead[4] = 0x24;
    thead[5] = 0x1d;
    thead[6] = 0x96;

    printf("##############################################################################\n");
    printf("#                                                                            #\n");
    printf("#  AX81b Filesystem Installer V1.26                                          #\n");
    printf("#  (c) 2012 Joerg Wolfram                                                    #\n");
    printf("#                                                                            #\n");
    if (argc < 5) {
        printf("#  Parameter: aaa bbb ccc ddd                                                #\n");
        printf("#                                                                            #\n");
        printf("#  aaa = Number of ZX81 Tapes (0 for skip)                                   #\n");
        printf("#  bbb = Number of ZX80 Tapes (0 for skip)                                   #\n");
        printf("#  ccc = Number of Jupiter ACE Tapes (0 for skip)                            #\n");
        printf("#  ddd = Number of ZX Spectrum Tapes (0 for skip)                            #\n");
        printf("#                                                                            #\n");
        printf("##############################################################################\n");
        printf("\nTOO FEW PARAMETERS!!\n\n");
        free(tprog);
        free(thead);
        return 1;
    } else {
        for (i = 0; i < 4; i++) {
            maxtapes[i] = atoi(argv[i + 1]);
        }
        printf("#  %03d  ZX81 Tapes (0 for skip)                                              #\n", maxtapes[0]);
        printf("#  %03d  ZX80 Tapes (0 for skip)                                              #\n", maxtapes[1]);
        printf("#  %03d  Jupiter ACE Tapes (0 for skip)                                       #\n", maxtapes[2]);
        printf("#  %03d  ZX Spectrum Tapes (0 for skip)                                       #\n", maxtapes[3]);
        printf("#                                                                            #\n");
        printf("##############################################################################\n");
    }

    for (this_image = 0; this_image < 4; this_image++) {
        if (maxtapes[this_image] > 0) {
            switch (this_image) {
            case 0:
                printf("Installing file system structure for ZX81...\n");
                thead[7] = 'Z';
                thead[8] = 'X';
                thead[9] = '8';
                thead[10] = '1';
                sprintf(filetype, "ZX81");
		imagefile_name = "ax81_x81.bin";
                break;
            case 1:
                printf("Installing file system structure for ZX80...\n");
                thead[7] = 'Z';
                thead[8] = 'X';
                thead[9] = '8';
                thead[10] = '0';
                sprintf(filetype, "ZX80");
		imagefile_name = "ax81_x80.bin";
                break;
            case 2:
                printf
                ("Installing file system structure for Jupiter ACE...\n");
                thead[7] = 'J';
                thead[8] = 'A';
                thead[9] = 'C';
                thead[10] = 'E';
                sprintf(filetype, "JACE");
		imagefile_name = "ax81_ace.bin";
                break;
            case 3:
                printf
                ("Installing file system structure for ZX Spectrum...\n");
                thead[7] = 'S';
                thead[8] = 'P';
                thead[9] = 'E';
                thead[10] = 'C';
                sprintf(filetype, "SPEC");
		imagefile_name = "ax81_spc.bin";
                break;
            default:
                break;
            }

            if (imagefile_name == NULL) {
                printf("unknown image type\n");
                free(tprog);
                free(thead);
                return 1;
            }

            imagefile = fopen(imagefile_name, "wb");
	    if (imagefile == NULL) {
                printf("cannot create image file %s\n", imagefile_name);
                free(tprog);
                free(thead);
                return 1;
	    }

            //add header sector for image size
            thead[11] = maxtapes[this_image] - 1;
            fwrite(thead, 1, 512, imagefile);
            for (tape = 0; tape < maxtapes[this_image]; tape++) {
                fwrite(tprog, 1, TAPE_FILE_LENGTH, imagefile);
            }
            fclose(imagefile);
            sprintf(filename, "TAPES81_%s", filetype);
            rstat = mkdir(filename, DIRECTORY_MODE);
            if (rstat == 0) {
                for (tape = 0; tape < maxtapes[this_image]; tape++) {
                    sprintf(filename, "TAPES81_%s/TAPE_%03u", filetype, tape);
                    rstat = mkdir(filename, DIRECTORY_MODE);
                }
            }
        }
    }
    printf("\nDONE.\n\n");
    free(tprog);
    free(thead);
    return 0;
}
