//###############################################################################
//#                                                                             #
//# AX81 computer with ATmega1284P                                              #
//# copyright (c) 2010-2012 Joerg Wolfram (joerg@jcwolfram.de)                  #
//#                                                                             #
//#                                                                             #
//# This program is free software; you can redistribute it and/or               #
//# modify it under the terms of the GNU General Public License as              #
//# published by the Free Software Foundation; either version 3 of              #
//# the License, or (at your option) any later version.                         #
//#                                                                             #
//# This library is distributed in the hope that it will be useful,             #
//# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
//# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU            #
//# General Public License for more details.                                    #
//#                                                                             #
//# You should have received a copy of the GNU General Public                   #
//# License along with this library; if not, write to the                       #
//# Free Software Foundation, Inc., 59 Temple Place - Suite 330,                #
//# Boston, MA 02111-1307, USA.                                                 #
//#                                                                             #
//###############################################################################

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>

#include "ax81-tools.h"

struct stat buf;


int main(void)
{
    char *buffer;
    //struct dirent *CurrentFile;
    //DIR * Directory;
    int tape, i, j, stat, tfile, w_char, nsize;
    unsigned char tprog[TAPE_WITH_HEADER_LENGTH];
    char filename[1000];
    FILE * imagefile;
    FILE * pfile;
    long offset;
    long filesize;
    //int size_low;
    //int size_high;
    emu_type_t fs_type;

    buffer = (char *) malloc(1000);

    if ((buffer = getcwd(NULL, 0)) == NULL) {
        perror("cannot get working dir");
        return EXIT_FAILURE;
    }

    tape = 0;
    fs_type = EMU_UNKNOWN;

    for (i = 0; i < 256; i++) {
        sprintf(filename, "/TAPES81_ZX81/TAPE_%03u", i);
        if (strstr(buffer, filename)) {
            tape = i;          //found
            fs_type = EMU_ZX81;
        }

        sprintf(filename, "/TAPES81_ZX80/TAPE_%03u", i);
        if (strstr(buffer, filename)) {
            tape = i;          //found
            fs_type = EMU_ZX80;
        }
        sprintf(filename, "/TAPES81_JACE/TAPE_%03u", i);
        if (strstr(buffer, filename)) {
            tape = i;          //found
            fs_type = EMU_ACE;
        }
        sprintf(filename, "/TAPES81_SPEC/TAPE_%03u", i);
        if (strstr(buffer, filename)) {
            tape = i;          //found
            fs_type = EMU_ZXSPECTRUM;
        }
    }
    printf("##############################################################################\n");
    printf("#                                                                            #\n");
    printf("#  AX81b Filesystem Reader V1.26 for Windows                                 #\n");
    printf("#  (c) 2012 Joerg Wolfram                                                    #\n");
    printf("#                                                                            #\n");

    switch (fs_type) {
    case EMU_ZX81:
        printf("#  ZX81 - Mode                                                               #\n");
        break;
    case EMU_ZX80:
        printf("#  ZX80 - Mode                                                               #\n");
        break;
    case EMU_ACE:
        printf("#  Jupiter ACE - Mode                                                        #\n");
        break;
    case EMU_ZXSPECTRUM:
        printf("#  ZX Spectrum - Mode                                                        #\n");
        break;
    default:
        printf("#  Unknown Mode                                                              #\n");
        printf("%s\n",filename);
        break;
    }
    printf("#                                                                            #\n");
    printf("##############################################################################\n\n");
    if (fs_type < EMU_ZX81) {

//              printf("WD=%s\n",buffer);
//              printf("FN=%s\n",filename);
        free(buffer);
        printf("Wrong place.\n");
        return EXIT_FAILURE;
    }

    memset(tprog, 255, TAPE_WITH_HEADER_LENGTH);

    offset = TAPE_FILE_HEADER_LENGTH + TAPE_FILE_LENGTH * tape;
    if (fs_type == EMU_ZX81) {
        printf("ZX81 Mode, Tape %d \n", tape);
        for (tfile = 0; tfile < TAPES_PER_FILE; tfile++) {
            imagefile = fopen("../../ax81_x81.bin", "rb");
            if (imagefile == NULL) {
                printf("error opening image file\n");
                free(buffer);
                return EXIT_FAILURE;
            }
            offset = TAPE_FILE_HEADER_LENGTH + TAPE_FILE_LENGTH * tape + TAPE_WITH_HEADER_LENGTH * tfile;
            stat = fseek(imagefile, offset, SEEK_SET);
            if (stat != 0) {
                perror("seek error");
                return EXIT_FAILURE;
            }
            
            stat = fread(tprog, 1, TAPE_WITH_HEADER_LENGTH, imagefile);
            if (stat <= 0) {
                perror("read error");
                return EXIT_FAILURE;
            }
            fclose(imagefile);
            if (tprog[0] < 128) {
                filesize = tprog[10] + 256 * tprog[11];
                buffer[0] = 0;
                nsize = 0;
                for (j = 0; j < 10; j++) {
                    if (tprog[j] > 32) {
                        w_char = tprog[j];
                        buffer[nsize++] = w_char;
                        buffer[nsize] = 0;
                    }
                }
                strcat(buffer, ".P");
                pfile = fopen(buffer, "wb");
                stat = fwrite(tprog + 512, 1, filesize, pfile);
                fclose(pfile);
            }
        }
    }

    if (fs_type == EMU_ZX80) {
        printf("ZX80 Mode, Tape %d \n", tape);
        for (tfile = 0; tfile < TAPES_PER_FILE; tfile++) {
            imagefile = fopen("../../ax81_x80.bin", "rb");
            if (imagefile == NULL) {
                printf("error opening image file\n");
                free(buffer);
                return EXIT_FAILURE;
            }
            offset = TAPE_FILE_HEADER_LENGTH + TAPE_FILE_LENGTH * tape + TAPE_WITH_HEADER_LENGTH * tfile;
            stat = fseek(imagefile, offset, SEEK_SET);
            stat = fread(tprog, 1, TAPE_WITH_HEADER_LENGTH, imagefile);
            fclose(imagefile);
            if (tprog[0] < 128) {
                filesize = tprog[10] + 256 * tprog[11];
                buffer[0] = 0;
                nsize = 0;
                for (j = 0; j < 10; j++) {
                    if (tprog[j] > 32) {
                        w_char = tprog[j];
                        buffer[nsize++] = w_char;
                        buffer[nsize] = 0;
                    }
                }
                strcat(buffer, ".O");
                pfile = fopen(buffer, "wb");
                stat = fwrite(tprog + 512, 1, filesize, pfile);
                fclose(pfile);
            }
        }
    }

    if (fs_type == EMU_ACE) {
        printf("Jupiter ACE Mode, Tape %d \n", tape);
        for (tfile = 0; tfile < TAPES_PER_FILE; tfile++) {
            imagefile = fopen("../../ax81_ace.bin", "rb");
            if (imagefile == NULL) {
                printf("error opening image file\n");
                free(buffer);
                return EXIT_FAILURE;
            }
            offset = TAPE_FILE_HEADER_LENGTH + TAPE_FILE_LENGTH * tape + TAPE_WITH_HEADER_LENGTH * tfile;
            stat = fseek(imagefile, offset, SEEK_SET);
            stat = fread(tprog, 1, TAPE_WITH_HEADER_LENGTH, imagefile);
            fclose(imagefile);
            
            if (tprog[0] < 128) {
                filesize = tprog[10] + 256 * tprog[11];
                unsigned char *buffer2 = malloc(filesize * 3);
                unsigned char *buffer1 = malloc(filesize < TAPE_LENGTH ? TAPE_LENGTH : filesize);
                memset(buffer2, 0, filesize*3);
                memset(buffer1, 0, filesize < TAPE_LENGTH ? TAPE_LENGTH : filesize);
                
                buffer[0] = 0;
                nsize = 0;
                for (j = 0; j < 10; j++) {
                    if (tprog[j] > 32) {
                        w_char = tprog[j];
                        buffer[nsize++] = w_char;
                        buffer[nsize] = 0;
                    }
                }
                printf("%s filesize %ld\n",buffer,filesize);

                unsigned char *rptr = tprog + TAPE_HEADER_LENGTH;
                unsigned char *wptr = buffer1;
                for (i = 0; i < TAPE_LENGTH; i++) {
                    *wptr++ = *rptr++;
                }
                
                rptr = buffer1;
                wptr = buffer2;
                for (i = 0; i < filesize; i++) {
                  if (((rptr[0] == rptr[1]) && (rptr[0] == rptr[2]))
                      || (rptr[0] == 0xed)) {
                    *wptr++ = 0xed;
                    unsigned char value = rptr[0];
                    int num = 0;
                    while ((*rptr++ == value) && (i < filesize) && ( num < 240)) {
                      i++;
                      num++;
                    }
                    *wptr++ = num;
                    *wptr++ = value;
                    rptr--;
                  } else {
                    *wptr++ = *rptr++;
                  }
                }
                *wptr++ = 0xed;
                *wptr++ = 0x00;
                strcat(buffer, ".ace");
                pfile = fopen(buffer, "wb");
                stat = fwrite(buffer2, 1, (wptr - buffer2), pfile);
                fclose(pfile);
                free(buffer1);
                free(buffer2);
            }
        }
    }

    if (fs_type == EMU_ZXSPECTRUM) {
        printf("ZX Spectrum Mode, Tape %d \n", tape);
        for (tfile = 0; tfile < TAPES_PER_FILE; tfile++) {
            imagefile = fopen("../../ax81_spc.bin", "rb");
            if (imagefile == NULL) {
                printf("error opening image file\n");
                free(buffer);
                return EXIT_FAILURE;
            }
            offset = TAPE_FILE_HEADER_LENGTH + TAPE_FILE_LENGTH * tape + TAPE_WITH_HEADER_LENGTH * tfile;
            stat = fseek(imagefile, offset, SEEK_SET);
            stat = fread(tprog, 1, TAPE_WITH_HEADER_LENGTH, imagefile);
            fclose(imagefile);
            if (tprog[0] < 128) {
#if 0
                filesize = tprog[10] + 256 * tprog[11];
#else
                filesize = 15360;
#endif
                unsigned char *buffer2 = malloc(65536);
                memset(buffer2, 0, 65536);
                
                buffer[0] = 0;
                nsize = 0;
                for (j = 0; j < 10; j++) {
                    if (tprog[j] > 32) {
                        w_char = tprog[j];
                        buffer[nsize++] = w_char;
                        buffer[nsize] = 0;
                    }
                }

                strcat(buffer, ".z80");
                pfile = fopen(buffer, "wb");

                unsigned char *rptr = tprog + 256;
                unsigned char *wptr = buffer2;
                for (i = 0; i < 30; i++) {
                     *wptr++ = *rptr++;
                }
                // write uncompressed mode1 file
                buffer2[12] &= ~(1 << 5);
                // write mode1 snapshot header
                stat = fwrite(buffer2, 1, 30, pfile);

                rptr = tprog + TAPE_HEADER_LENGTH;
                wptr = buffer2;
                for (i = 0; i < filesize; i++) {
                     *wptr++ = *rptr++;
                }
                
#if 0
                int filesize = TAPE_LENGTH;
                unsigned char pagehead[3];
                // page header length set to 0xffff (16384 byte uncompressed data)
                pagehead[0] = 0xff;
                pagehead[1] = 0xff;
                pagehead[2] = 4;
                int offset = 16384;
                stat = fwrite(pagehead, 1, 3, pfile);
                stat = fwrite(buffer2 + offset, 1, 16384, pfile);

                pagehead[2] = 5;
                offset = 32768;
                stat = fwrite(pagehead, 1, 3, pfile);
                stat = fwrite(buffer2 + offset, 1, 16384, pfile);

                pagehead[2] = 8;
                offset = 0;
                stat = fwrite(pagehead, 1, 3, pfile);
                stat = fwrite(buffer2 + offset, 1, 16384, pfile);
#else
                stat = fwrite(buffer2, 1, 49152, pfile);
#endif
                fclose(pfile);
                free(buffer2);
            }
        }
    }
    free(buffer);
    return (0);
}

